open Viz_types

(* find the connected components of the ancestry subgraph. *)
let components subgraph =
  let already_seen comp_list id =
    List.exists (NodeMap.mem id) comp_list in

  let rec add_to_comp comp = function
    | [] ->
	comp
    | node :: tl when NodeMap.mem node.id comp ->
	add_to_comp comp tl
    | node :: tl ->
	let comp = NodeMap.add node.id node comp in
	let tl =
	  List.fold_left
	    (fun acc (rel_id, _) ->
	      if NodeMap.mem rel_id comp
	      then acc
	      else 
		try (NodeMap.find rel_id subgraph) :: acc
		with Not_found -> acc)
	    tl
	    node.family in
	add_to_comp comp tl in

  NodeMap.fold
    (fun id node acc ->
      if already_seen acc id
      then acc
      else (add_to_comp NodeMap.empty [ node ]) :: acc)
    subgraph
    []


(* A not-too-dumb DFS (ie tail-recursive) *)

type 'a dfs_data =
  | Nil 
  | Node  of 'a * 'a dfs_data
  | Child of 'a * 'a dfs_data

let rec explore get_children f ((explored, f_acc) as acc) = function
  | Nil -> 
      acc
  | Node (node, tl) 
  | Child (node, tl) when IdSet.mem node.id explored ->
      explore get_children f acc tl
  | Node (node, tl) ->
      explore get_children f
	(IdSet.add node.id explored, f node f_acc)
	tl
  | Child (child, tl) ->
      explore get_children f
	acc
	(get_children child 
	   (fun acc child -> Child (child, acc))
	   (Node (child, tl)))
	    
let dfs subgraph get_children f init =
  let (_, res) =
    NodeMap.fold
      (fun _ node acc -> 
	if List.for_all (function (_, PARENT) -> false | (_, CHILD) -> true) node.family
	then explore get_children f acc (Child (node, Nil))
	else acc)
      subgraph 
      (IdSet.empty, init) in
  res


let fold_children g node f init =
  List.fold_left
    (fun acc -> function
      | (_, PARENT) -> 
	  acc
      | (rel_id, CHILD) ->
	  try f acc (NodeMap.find rel_id g)
	  with Not_found -> acc)
      init node.family

let topo_sort_neighbors subgraph =
  List.rev
    (dfs
       subgraph
       (fold_children subgraph)
       (fun node acc ->
	 if node.kind = NEIGHBOUR_IN || node.kind = NEIGHBOUR_OUT
	 then node :: acc
	 else acc)
       [])
    

let rec explore_bis get_children f explored q = 
  match Q.pop q with
  | None ->
      (None, explored)
  | Some (node, tl) when IdSet.mem node explored ->
      explore_bis get_children f explored tl
  | Some (node, tl) ->
      let explored = IdSet.add node explored in
      match f node with
      | `ACCEPT -> 
	  (Some node, explored)
      | `REJECT -> 
	  explore_bis get_children f explored tl
      | `CONTINUE ->
	  explore_bis get_children f explored
	    (get_children node Q.push tl)

let rec apply_till_found f acc = function
  | [] -> None
  | h :: tl ->
      match f acc h with
      | (None, acc) ->
	  apply_till_found f acc tl
      | (v, _) -> v


let reconnect fetch_children agraph =
  match components agraph.nodes with
  | [] | [_] -> 
      Viz_misc.log "comp" "connected graph" ;
      agraph

  | comps ->
      Viz_misc.log "comp" "%d components" (List.length comps) ;

      let comps_with_neighbors =
	List.map (fun c -> c, topo_sort_neighbors c) comps in

      if Viz_misc.debug "comp" then begin
	List.iter
	  (fun (_, n_sorted) ->
	    Viz_misc.log "comp" "topo sort of component neighbors (%d):\n  %s"
	      (List.length n_sorted)
	      (String.concat "\n  " 
		 (List.map (fun node -> node.id) n_sorted)))
	  comps_with_neighbors
      end ;

      let edges =
	List.fold_left
	  (fun acc (comp, neighbors_sorted) ->
	    (* for each component, try to connect it to
	       at most one other component. *)

	    let opt_edge =
	      apply_till_found
		(fun explored start_neighbor ->
		  match
		    explore_bis fetch_children
		      (fun node_id ->
			try 
			  let node = NodeMap.find node_id agraph.nodes in
			  if NodeMap.mem node_id comp
			  then `REJECT
			  else if node.kind = NEIGHBOUR_IN
			  then `ACCEPT
			  else begin
			    assert (node.kind = NEIGHBOUR_OUT) ;
			    `CONTINUE
			  end
			with Not_found -> `CONTINUE)
		      explored
		      (fetch_children start_neighbor.id Q.push Q.empty)
		  with
		  | (Some target, e) -> 
		      Viz_misc.log "comp"
			"found an edge: %s -> %s" start_neighbor.id target ;
		      (Some (start_neighbor.id, target), e)
		  | (None, _) as r -> r)
		IdSet.empty
		neighbors_sorted in
	    
	    match opt_edge with
	    | None -> acc
	    | Some edge -> 
		EdgeMap.add edge SPANNING acc)

	  agraph.ancestry

	  comps_with_neighbors in
      
      { agraph with ancestry = edges }
