open Viz_misc

type mt_options =
  | MTopt_none
  | MTopt_db of string
  | MTopt_full of string * string

let unquote s =
  if s.[0] = '"'
  then 
    let len = String.length s in
    Revision_lexer.string 
      (Buffer.create len)
      (Lexing.from_string (String.sub s 1 (len - 1)))
  else s

let find_MT_dir () =
  let rec up = function
    | "/" -> raise Not_found
    | p ->
	let d = Filename.dirname p in
	let m = Filename.concat d "MT" in
	if Sys.file_exists m 
	then m
	else up d in
  if Sys.file_exists "MT"
  then "MT"
  else up (Sys.getcwd ())

let parse_MT_options () =
  match
    try with_file_in input_lines (Filename.concat (find_MT_dir ()) "options")
    with Not_found | Sys_error _ -> [] with
  | [] -> MTopt_none
  | lines ->
      let options = 
	try
	  List.fold_right
	    (fun s acc ->
	      match string_split ~max_elem:2 ' ' s with
	      | [a; b] -> (a, unquote b) :: acc
	      | _ -> acc)
	    lines [] 
	with Failure _ -> [] in
      match may_assoc "database" options with
      | None -> MTopt_none
      | Some db_raw ->
	  try
	    let db = Glib.Convert.filename_from_utf8 db_raw in
	    match may_assoc "branch" options with
	    | Some branch when Glib.Utf8.validate branch -> 
		MTopt_full (db, branch)
	    | _ -> 
		MTopt_db db
	  with Glib.Convert.Error _ ->
	    MTopt_none

let parse_options args =
  match args with
  | [] -> 
      parse_MT_options ()
  | db :: [] | db :: "" :: _ -> 
      MTopt_db db
  | db :: branch_raw :: _ -> 
      try
	let branch = Glib.Convert.locale_to_utf8 branch_raw in
	MTopt_full (db, branch)
      with Glib.Convert.Error _ ->
	MTopt_db db

let parse_cli () =
  let anons = ref Q.empty in
  let aa = ref true in
  let cli_args = [ "-noaa", Arg.Clear aa, "don't use an anti-aliased canvas" ] in
  let usg_msg = 
    Printf.sprintf "usage: %s [options] [db [branch]]"
      (Filename.basename Sys.executable_name) in
  Arg.parse cli_args (fun a -> anons := Q.push !anons a) usg_msg ;
  (!aa, parse_options (Q.to_list !anons))


let exn_handler w exn = 
  View.error_notice ~parent:w 
    (match exn with
    | Viz_types.Error msg -> msg
    | exn -> Printexc.to_string exn)

let main =
  let w = GWindow.window 
      ~title:"Monotone-viz"
      ~icon:(Lazy.force Icon.monotone) () in
  ignore (w#connect#destroy GMain.quit) ;

  let (aa, mt_options) = parse_cli () in

  GtkSignal.user_handler := exn_handler w ;

  let prefs = Viz_style.load () in

  let v = Ui.get_view (Ui.make w ~aa ~prefs) in

  ignore 
    (Glib.Idle.add 
       (fun () -> 
	 begin
	   try
	     match mt_options with
	     | MTopt_none -> ()
	     | MTopt_db fname ->
		 View.open_db v fname None
	     | MTopt_full (fname, branch) ->
		 View.open_db v fname (Some branch)
	   with Viz_types.Error msg -> 
	     View.error_notice ~parent:w msg
	 end ;
	 false)) ; 

  w#show () ;
  GMain.main () ;

  (* just close the db, without updating the widgets *)
  View.finalize v
