#include <string.h>
#include <assert.h>

#include <caml/alloc.h>
#include <caml/memory.h>
#include <caml/fail.h>
#include <caml/callback.h>
#include <caml/custom.h>

#include <sqlite3.h>

#define TRUE  1
#define FALSE 0

struct user_function {
  value fun;
  struct user_function *next;
};

struct ml_sqlite3_data {
  sqlite3 *db;
  value  callbacks;
  struct user_function *user_functions;
};
/* 0 -> trace
 * 1 -> busy
 * 2 -> progress
 */
#define NUM_CALLBACKS 3
#define Sqlite3_val(v) (((struct ml_sqlite3_data *) Data_custom_val(v))->db)

static void 
ml_sqlite3_finalize (value v)
{
  struct user_function *list, *next;;
  struct ml_sqlite3_data *data = Data_custom_val(v);
  remove_global_root (&data->callbacks);
  list = data->user_functions;
  while (list != NULL)
    {
      remove_global_root (&list->fun);
      next = list->next;
      stat_free (list);
      list = next;
    }
}

static value
ml_wrap_sqlite3 (sqlite3 *db)
{
  static struct custom_operations ops = {
    "mlsqlite3/001", 
    ml_sqlite3_finalize 
  };
  struct ml_sqlite3_data *data;
  value v;
  v = alloc_custom (&ops, sizeof (struct ml_sqlite3_data), 1, 100);
  data = Data_custom_val (v);
  data->db = db;
  data->callbacks = alloc (NUM_CALLBACKS, 0);
  data->user_functions = NULL;
  register_global_root (&data->callbacks);
  return v;
}


static void ml_sqlite3_raise_exn (int, char *, int) Noreturn;
static void ml_sqlite3_raise_exn (int status, char *errmsg, int static_errmsg)
{
  static value *sqlite3_exn;

  /* check this is really an error */
  assert (status > SQLITE_OK && status < SQLITE_ROW);

  if (sqlite3_exn == NULL)
    {
      sqlite3_exn = caml_named_value ("mlsqlite3_exn");
      if (sqlite3_exn == NULL)
	failwith ("Sqlite3 exception not registered");
    }

  { 
    CAMLlocal1(bucket);
    bucket = alloc (3, 0);
    Store_field(bucket, 0, *sqlite3_exn);
    Store_field(bucket, 1, Val_long (status - 1));
    Store_field(bucket, 2, copy_string (errmsg ? errmsg : ""));
    if (!static_errmsg)
      sqlite3_free (errmsg);
    mlraise (bucket);
  }
}

CAMLprim value ml_sqlite3_open (value filename)
{
  sqlite3 *db;
  int status;

  status = sqlite3_open (String_val(filename), &db);
  if (status != SQLITE_OK)
    {
      char *errmsg;
      errmsg = (char *) sqlite3_errmsg (db);
      sqlite3_close (db); /* ignore status here */
      ml_sqlite3_raise_exn (status, errmsg, TRUE);
    }

  return ml_wrap_sqlite3 (db);
}

CAMLprim value ml_sqlite3_close (value db)
{
  int status;
  status = sqlite3_close (Sqlite3_val(db));
  if (status != SQLITE_OK)
    {
      char *errmsg;
      errmsg = (char *) sqlite3_errmsg (Sqlite3_val (db));
      ml_sqlite3_raise_exn (status, errmsg, TRUE);
    }
  Sqlite3_val(db) = NULL;
  return Val_unit;
}

struct cb_data {
  value *closure;
  value *accum;
  value *column_names_array;
  int   option_string;
  enum {
    COLUMN_NAMES_NEED_STORE,
    COLUMN_NAMES_STORED
  } columns;
};

static value 
copy_opt_string (char *s)
{
  if (s == NULL)
    return Val_unit;
  else
    {
      CAMLparam0();
      CAMLlocal2(str, b);

      str = copy_string(s);
      b = alloc_small(1, 0);
      Field(b, 0) = str;
      CAMLreturn(b);
    }
}

static value 
ml_copy_string_array (int len, char **s_arr, int option_string)
{
  unsigned int i;
  CAMLparam0();
  CAMLlocal2(arr, tmp);
  
  if (len <= 0 || s_arr == NULL)
    CAMLreturn (Atom(0));

  arr = alloc (len, 0);
  for (i=0; i<len; i++)
    {
      if (option_string)
	tmp = copy_opt_string (s_arr[i]);
      else if (s_arr[i] == NULL)
	tmp = copy_string ("");
      else
	tmp = copy_string (s_arr[i]);
      Store_field (arr, i, tmp);
    }
  CAMLreturn (arr);
}

static int ml_sqlite3_callback (void *data, int argc, char **argv, char **col_names)
{
  CAMLparam0();
  CAMLlocal2(res, values);
  struct cb_data *cb_data = data;

  if (cb_data->columns == COLUMN_NAMES_NEED_STORE && 
      Is_block (*cb_data->column_names_array))
    {
      CAMLlocal2 (ref, ml_col_names);
      /* column_names_array if of type `array ref option' */
      ref = Field (*cb_data->column_names_array, 0);
      ml_col_names = ml_copy_string_array (argc, col_names, 0);
      Store_field (ref, 0, ml_col_names);
      cb_data->columns = COLUMN_NAMES_STORED;
    }

  values = ml_copy_string_array (argc, argv, cb_data->option_string);

  res = callback2_exn (*cb_data->closure, *cb_data->accum, values);
  *cb_data->accum = res;
  if (Is_exception_result (res))
    CAMLreturn (SQLITE_ERROR);

  CAMLreturn (SQLITE_OK);
}

static value 
ml_sqlite3_fetch_gen (value db, value sql, 
		      value o_col_names, 
		      value cb, value init, int opt_string)
{
  CAMLparam2 (cb, o_col_names);
  CAMLlocal1 (accum);
  struct cb_data cb_data = { &cb, &accum, &o_col_names, 
			     opt_string, COLUMN_NAMES_NEED_STORE };
  char *sql_copy, *errmsg;
  int status;

  errmsg = NULL;
  accum = init;
  sql_copy = stat_alloc (string_length (sql) + 1);
  strcpy (sql_copy, String_val (sql));
  status = sqlite3_exec (Sqlite3_val (db), sql_copy,
			 ml_sqlite3_callback, &cb_data, &errmsg);
  stat_free (sql_copy);
  if (status != SQLITE_OK)
    {
      if (Is_exception_result (accum)) 
	{
	  sqlite3_free (errmsg); /* should be NULL actually */
	  mlraise (Extract_exception (accum));
	}
      else
	ml_sqlite3_raise_exn (status, errmsg, FALSE);
    }
  CAMLreturn(accum);
}

CAMLprim value 
ml_sqlite3_fetch (value db, value sql, 
		  value o_col_names, 
		  value cb, value init)
{
  return ml_sqlite3_fetch_gen (db, sql, o_col_names, cb, init, FALSE);
}

CAMLprim value 
ml_sqlite3_fetch_opt (value db, value sql, 
		      value o_col_names, 
		      value cb, value init)
{
  return ml_sqlite3_fetch_gen (db, sql, o_col_names, cb, init, TRUE);
}

CAMLprim value 
ml_sqlite3_exec (value db, value sql)
{
  char *sql_copy, *errmsg;
  int status;

  sql_copy = stat_alloc (string_length (sql) + 1);
  strcpy (sql_copy, String_val (sql));
  status = sqlite3_exec (Sqlite3_val (db), String_val (sql), 
			 NULL, NULL, &errmsg);
  stat_free (sql_copy);
  if (status != SQLITE_OK)
    ml_sqlite3_raise_exn (status, errmsg, FALSE);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_interrupt (value db)
{
  sqlite3_interrupt (Sqlite3_val (db));
  return Val_unit;
}

CAMLprim value ml_sqlite3_complete (value sql)
{
  return Val_bool (sqlite3_complete (String_val (sql)));
}

#define MLTAG_RETRY 0xc96e9e91
#define MLTAG_FAIL  0x5ced03bd

static int
ml_sqlite3_busy_handler_cb (void *data, int num)
{
  struct ml_sqlite3_data *db = data;
  value res;
  res = callback_exn (Field (db->callbacks, 0), Val_int (num));
  if (Is_exception_result (res))
    return 0;
  if (res == MLTAG_RETRY)
    return 1;
  else
    return 0;
}

CAMLprim value
ml_sqlite3_busy_handler (value db, value cb)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 0, cb);
  sqlite3_busy_handler (Sqlite3_val(db), 
			ml_sqlite3_busy_handler_cb, db_data);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_busy_handler_unset (value db)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 0, Val_unit);
  sqlite3_busy_handler (Sqlite3_val(db), NULL, NULL);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_busy_timeout (value db, value t)
{
  sqlite3_busy_timeout (Sqlite3_val (db), Int_val (t));
  return Val_unit;
}

static void 
ml_sqlite3_trace_handler (void *data, const char *req)
{
 struct ml_sqlite3_data *db = data;
 value s = copy_string (req);
 callback_exn (Field (db->callbacks, 1), s);
}

CAMLprim value 
ml_sqlite3_trace (value db, value cb)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 1, cb);
  sqlite3_trace (Sqlite3_val (db), ml_sqlite3_trace_handler, db_data);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_trace_unset (value db)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 1, Val_unit);
  sqlite3_trace (Sqlite3_val (db), NULL, NULL);
  return Val_unit;
}

static int 
ml_sqlite3_progress_handler_cb (void *data)
{
 struct ml_sqlite3_data *db = data;
 value res;
 res = callback_exn (Field (db->callbacks, 2), Val_unit);
 return Is_exception_result(res);
}

CAMLprim value 
ml_sqlite3_progress_handler (value db, value delay, value cb)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 2, cb);
  sqlite3_progress_handler (Sqlite3_val (db), Int_val (delay), 
			    ml_sqlite3_progress_handler_cb, db_data);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_progress_handler_unset (value db)
{
  struct ml_sqlite3_data *db_data = Data_custom_val (db);
  Store_field (db_data->callbacks, 2, Val_unit);
  sqlite3_progress_handler (Sqlite3_val(db), 0, NULL, NULL);
  return Val_unit;
}

CAMLprim value 
ml_sqlite3_version (value unit)
{
  return copy_string (sqlite3_version);
}

CAMLprim value
ml_sqlite3_last_insert_rowid (value db)
{
  return copy_int64 (sqlite3_last_insert_rowid (Sqlite3_val (db)));
}

CAMLprim value
ml_sqlite3_changes (value db)
{
  return Val_long (sqlite3_changes (Sqlite3_val (db)));
}

CAMLprim value
ml_sqlite3_total_changes (value db)
{
  return Val_long (sqlite3_total_changes (Sqlite3_val (db)));
}


/* User-defined functions */

static struct user_function *
register_user_function (struct ml_sqlite3_data *db_data, value name, value cb)
{
  CAMLparam2(name, cb);
  CAMLlocal1(cell);
  struct user_function *link;
  cell = alloc (2, 0);
  Store_field (cell, 0, name);
  Store_field (cell, 1, cb);
  link = stat_alloc (sizeof *link);
  link->fun = cell;
  link->next = db_data->user_functions;
  register_global_root (&link->fun);
  db_data->user_functions = link;
  CAMLreturn(link);
}

static void
unregister_user_function (struct ml_sqlite3_data *db_data, value name)
{
  struct user_function *prev, *link;
  prev = NULL;
  link = db_data->user_functions;
  while (link != NULL)
    {
      if (strcmp (String_val (Field (link->fun, 0)), String_val (name)) == 0)
	{
	  if (prev == NULL)
	    db_data->user_functions = link->next;
	  else
	    prev->next = link->next;
	  remove_global_root (&link->fun);
	  stat_free (link);
	  break;
	}
      prev = link;
      link = link->next;
    }
}

static value
ml_sqlite3_wrap_values (int argc, sqlite3_value **args)
{
  int i;
  CAMLparam0();
  CAMLlocal2(a, v);
  if (argc == 0 || args == NULL)
    CAMLreturn (Atom (0));
  a = alloc (argc, 0);
  for (i=0; i<argc; i++)
    {
      v = alloc_small (1, Abstract_tag);
      Field (v, 0) = Val_bp (args[i]);
      Store_field (a, i, v);
    }
  CAMLreturn (a);
}

static void
ml_sqlite3_wipe_values (value a)
{
  mlsize_t i, len = Wosize_val (a);
  for (i=0; i<len; i++)
    Store_field (Field (a, i), 0, 0);
}

#define SQLITE3_VALUE \
  sqlite3_value *val = (sqlite3_value *) Field (v, 0); \
  if (val == NULL) invalid_argument ("sqlite3_value")

CAMLprim value
ml_sqlite3_value_blob (value v)
{
  value r;
  int len;
  const void *data;
  SQLITE3_VALUE;
  len = sqlite3_value_bytes (val);
  r = alloc_string (len);
  data = sqlite3_value_blob (val);
  memcpy (String_val (r), data, len);
  return r;
}

CAMLprim value
ml_sqlite3_value_double (value v)
{
  SQLITE3_VALUE;
  return copy_double (sqlite3_value_double (val));
}

CAMLprim value
ml_sqlite3_value_int (value v)
{
  SQLITE3_VALUE;
  return Val_long (sqlite3_value_int (val));
}

CAMLprim value
ml_sqlite3_value_int64 (value v)
{
  SQLITE3_VALUE;
  return copy_int64 (sqlite3_value_int64 (val));
}

CAMLprim value
ml_sqlite3_value_text (value v)
{
  SQLITE3_VALUE;
  return copy_string ((char *) sqlite3_value_text (val));
}

#define MLTAG_INTEGER 0x2ddf233d
#define MLTAG_FLOAT   0x0109faf9
#define MLTAG_TEXT    0x6f75295b
#define MLTAG_BLOB    0x57b40abb
#define MLTAG_NULL    0x679ecd0f

CAMLprim value
ml_sqlite3_value_type (value v)
{
  SQLITE3_VALUE;
  switch (sqlite3_value_type (val))
    {
    case SQLITE_INTEGER:
      return MLTAG_INTEGER;
    case SQLITE_FLOAT:
      return MLTAG_FLOAT;
    case SQLITE_TEXT:
      return MLTAG_TEXT;
    case SQLITE_BLOB:
      return MLTAG_BLOB;
    default:
      return MLTAG_NULL;
    }
}

#define MLTAG_INT     0x006f519f
#define MLTAG_INT64   0x781dd39b
#define MLTAG_VALUE   0x5f4d6ea3

static void 
ml_sqlite3_set_result (sqlite3_context *ctx, value res)
{
  if (Is_exception_result (res))
    sqlite3_result_error (ctx, "ocaml callback raised an exception", -1);
  else if (Is_long (res))
    sqlite3_result_null (ctx);
  else
    {
      value v = Field (res, 1);
      switch (Field (res, 0))
	{
	case MLTAG_INT:
	  sqlite3_result_int (ctx, Long_val (v));
	  break;
	case MLTAG_INT64:
	  sqlite3_result_int64 (ctx, Int64_val (v));
	  break;
	case MLTAG_FLOAT:
	  sqlite3_result_double (ctx, Double_val (v));
	  break;
	case MLTAG_TEXT:
	  sqlite3_result_text (ctx, String_val (v), string_length (v), SQLITE_TRANSIENT);
	  break;
	case MLTAG_BLOB:
	  sqlite3_result_blob (ctx, String_val (v), string_length (v), SQLITE_TRANSIENT);
	  break;
	case MLTAG_VALUE:
	  {
	    SQLITE3_VALUE;
	    sqlite3_result_value (ctx, val);
	    break;
	  }
	default:
	  sqlite3_result_error (ctx, "unknown value returned by callback", -1);
	}
    }
}

static void
ml_sqlite3_user_function (sqlite3_context *ctx, int argc, sqlite3_value **argv)
{
  struct user_function *data = sqlite3_user_data (ctx);
  CAMLparam0();
  CAMLlocal2(res, args);

  args = ml_sqlite3_wrap_values (argc, argv);
  res = callback_exn (Field (data->fun, 1), args);
  ml_sqlite3_set_result (ctx, res);
  ml_sqlite3_wipe_values (args);
  CAMLreturn0;
}

CAMLprim value
ml_sqlite3_create_function (value db, value name, value nargs, value fun)
{
  CAMLparam4(db, name, nargs, fun);
  int status;
  struct user_function *param = register_user_function (Data_custom_val (db), 
							name, fun);
  status = sqlite3_create_function (Sqlite3_val (db), String_val (name),
				    Int_val (nargs), SQLITE_UTF8, param,
				    ml_sqlite3_user_function, NULL, NULL);
  if (status != SQLITE_OK)
    {
      char *errmsg;
      errmsg = (char *) sqlite3_errmsg (Sqlite3_val (db));
      ml_sqlite3_raise_exn (status, errmsg, TRUE);
    }
  CAMLreturn(Val_unit);
}

CAMLprim value
ml_sqlite3_delete_function (value db, value name)
{
  int status;
  status = sqlite3_create_function (Sqlite3_val (db), String_val (name),
				    0, SQLITE_UTF8, NULL,
				    NULL, NULL, NULL);
  if (status != SQLITE_OK)
    {
      char *errmsg;
      errmsg = (char *) sqlite3_errmsg (Sqlite3_val (db));
      ml_sqlite3_raise_exn (status, errmsg, TRUE);
    }
  unregister_user_function (Data_custom_val (db), name);
  return Val_unit;
}
